/*
 * $Id$
 *
 * Serial Wire Debug Open API.
 * Library Body File.
 *
 * Copyright (C) 2010, Tomasz Boleslaw CEDRO (http://www.tomek.cedro.info)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of the Tomasz Boleslaw CEDRO nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.*
 *
 * Written by Tomasz Boleslaw CEDRO <tomek.cedro@gmail.com>, 2010;
 *
 */

#include <libswd.h>
#include <string.h>
#include <stdlib.h>

/* DATA GENERATED BELOW ARE TO BE SHIFTED OUT MSB FIRST, UNLIKE IN JTAG!
 * These functions generate a bistream packets for SWD communications.
 * Bistream is returned in byte array pointed by *buffer parameter.
 * Actual positive byte count is returned by the function.
 * Negative return value means an error. Zero means no data was generated.
 * If you need to swap bit direction, use bitswap functions provided below.
 * All functions name start with "swd_bitgen_" prefix.
 */

/* Even parity calculator helper function for int type. */
int swd_bin8_parity_even(char data){
 char i,parity=0;
 unsigned char test=data;
 for (i=0;i<=8;i++) parity ^= ((test>>i)&1);
 if (parity<0 || parity>1) return SWD_ERROR_PARITY;
 return parity;
}

int swd_bin32_parity_even(int data){
 int i,parity=0;
 unsigned int test=data;
 for (i=0;i<32;i++) parity ^= (test>>i)&1; 
 if (parity<0 || parity>1) return SWD_ERROR_PARITY;
 return parity;
}

int swd_bin8_print(char data){
 unsigned char i, bits=data;
 for (i=0;i<8;i++) putchar(((bits<<i)&0x80)?'1':'0'); 
 return i;
}

int swd_bin32_print(int data){
 unsigned int i, bits=data;
 for (i=0;i<32;i++) putchar(((bits<<i)&0x80000000)?'1':'0');
 return i;
}

char *swd_bin8_string(char data){
 static char string[9]; string[8]=0;
 unsigned char i, bits=data;
 for (i=0;i<8;i++) string[7-i]=(bits&(1<<i))?'1':'0'; 
 return string;
}

char *swd_bin32_string(int data){
 static char string[33]; string[32]=0;
 unsigned int i, bits=data;
 for (i=0;i<32;i++) string[31-i]=(bits&(1<<i))?'1':'0';
 return string;
}




/* BitSwap helper function that reverse bit order in *buffer. 
 * Least Significant Bit becomes Most Significant Bit.
 * Rersult contains only "count" bits for use with char or int easily.
 * This function operates on unsigned char (8-bit) data type.
 */
int swd_bin8_bitswap(unsigned char *buffer, int bitcount){
 if (bitcount>8) return SWD_ERROR_PARAM;
 unsigned char bit, res=0; //res must be unsigned for proper shifting result
 #ifdef __SWDDEBUG__
 printf("|SWD_DEBUG: swd_bin8_bitswap(%02X, %d);\n", *buffer, bitcount);
 #endif
 for (bit=0;bit<bitcount;bit++) {
  res=(res<<1)|(((*buffer>>bit)&1)?1:0); 
  #ifdef __SWDDEBUG__
  printf("|SWD_DEBUG: swd_bin8_bitswap: in=%02X out=%02X bit=%d\n", *buffer, res, bit);
  #endif
 }
 *buffer=res;
 return SWD_OK;
}

/* BitSwap helper function that reverse bit order in *buffer. 
 * Least Significant Bit becomes Most Significant Bit.
 * Rersult contains only "count" bits for use with char or int easily.
 * This function operates on unsigned int (32-bit) data type.
 */
int swd_bin32_bitswap(unsigned int *buffer, int bitcount){
 if (bitcount>32) return SWD_ERROR_PARAM;
 unsigned int bit, res=0; //res must be unsigned for proper shifting result
 #ifdef __SWDDEBUG__
 printf("|SWD_DEBUG: swd_bin32_bitswap(%08X, %d);\n", *buffer, bitcount);
 #endif
 for (bit=0;bit<bitcount;bit++) {
  res=(res<<1)|(((*buffer>>bit)&1)?1:0); 
  #ifdef __SWDDEBUG__
  printf("|SWD_DEBUG: swd_bin32_bitswap: in=%08X out=%08X bit=%d\n", *buffer, res, bit);
  #endif
 }
 *buffer=(int)res;
 return SWD_OK;
}




/* Command Queue helper functions.
 * All functions start with "swd_cmd_queue_" prefix. 
 */

/* Find first element of a queue pointed by *cmdq element. */
swd_cmd_t* swd_cmd_queue_find_root(swd_cmd_t *cmdq){
 if (cmdq==NULL) return NULL;
 swd_cmd_t *cmd=cmdq;
 while (cmd->prev!=NULL) cmd=cmd->prev;
 return cmd;
}

/* Find last element of a queue pointed by *cmdq element. */
swd_cmd_t* swd_cmd_queue_find_end(swd_cmd_t *cmdq){
 if (cmdq==NULL) return NULL;
 swd_cmd_t *cmd=cmdq;
 while (cmd->next!=NULL) cmd=cmd->next;
 return cmd;
}

/* Append element pointed by *cmd at the end of the quque pointed by *cmdq. */
int swd_cmd_queue_append(swd_cmd_t *cmdq, swd_cmd_t *cmd){
 if (cmdq==NULL || cmd==NULL) return SWD_ERROR_NULLPOINTER;
 if (cmdq->next != NULL){
  swd_cmd_t *lastcmd;
  lastcmd=swd_cmd_queue_find_end(cmdq);
  lastcmd->next=cmd;
  cmd->prev=lastcmd;
 } else {
  cmdq->next=cmd;
  cmd->prev=cmdq;
 }
 cmdq=cmd; 
 return 1;
}

/* Free memory / destroy queue pointed by the *cmdq element. */
int swd_cmd_queue_free(swd_cmd_t *cmdq){
 swd_cmd_t *cmd, *nextcmd;
 cmd=swd_cmd_queue_find_root(cmdq);
 while (cmd!=NULL) {
  nextcmd=cmd->next;
  free(cmd);
  cmd=nextcmd;
 }
 return SWD_OK;
}

/* Free memory / destroy / shorten queue head from root to the *cmdq element. */
int swd_cmd_queue_free_head(swd_cmd_t *cmdq){
 swd_cmd_t *cmdqroot, *nextcmd;
 cmdqroot=swd_cmd_queue_find_root(cmdq);
 while(cmdqroot!=cmdq){
  nextcmd=cmdqroot->next;
  free(cmdqroot);
  cmdqroot=nextcmd;
 }
 cmdqroot->prev=NULL;
 return SWD_OK;
}

int swd_cmd_queue_free_tail(swd_cmd_t *cmdq){
 swd_cmd_t *cmdqend, *nextcmd;
 nextcmd=cmdq;
 cmdqend=swd_cmd_queue_find_end(cmdq);
 if (cmdqend==NULL) return SWD_ERROR_QUEUE; 
 while(nextcmd!=cmdq){
  nextcmd=cmdqend->prev;
  free(cmdqend);
  cmdqend=nextcmd;
 }
 cmdqend->next=NULL;
 return SWD_OK;
}





/* SWD Commands Queue Elements Generator Routines.
 * These command quants are created in memory and can be easily appended
 * to the queue pointed by the *cmdq element.
 * All functions start with "swd_cmd_append_" prefix. */

/* Append SWD Request Packet Header to the context command queue.
 * Note that request is not validated here so bad request can be sent as well.*/
int swd_cmd_queue_append_request(swd_ctx_t *swdctx, char request){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
 swd_cmd_t *cmd;
 cmd=calloc(1,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 cmd->request=request;
 cmd->bits=SWD_REQUEST_BITLEN;
 cmd->cmdtype=SWD_CMDTYPE_MOSI;
 return swd_cmd_queue_append(swdctx->cmdq, cmd);
} 

/* Append SWD Turnaround to the context command queue. */ 
int swd_cmd_queue_append_mosi_trn(swd_ctx_t *swdctx){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
  swd_cmd_t *cmd;
  cmd=calloc(1,sizeof(swd_cmd_t));
  if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
  cmd->TRNnMOSI=0;
  cmd->bits=swdctx->params.trnlen;
  cmd->cmdtype=SWD_CMDTYPE_TRNMOSI;
  return swd_cmd_queue_append(swdctx->cmdq, cmd);
}

int swd_cmd_queue_append_miso_trn(swd_ctx_t *swdctx){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
  swd_cmd_t *cmd;
  cmd=calloc(1,sizeof(swd_cmd_t));
  if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
  cmd->TRNnMOSI=1;
  cmd->bits=swdctx->params.trnlen;
  cmd->cmdtype=SWD_CMDTYPE_TRNMISO;
  return swd_cmd_queue_append(swdctx->cmdq, cmd);
}



/* Append bus binary read bit-by-bit operation into context command queue.
 * Note that each command returns binary value into integer data element.
 * On error memory is freed and error code is returned.
 * Memory pointed by *data must be allocated prior calling this function. */
int swd_cmd_queue_append_miso_nbit(swd_ctx_t *swdctx, int *data, int count){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
 if (data==NULL) return SWD_ERROR_NULLPOINTER;
 if (count<=0) return SWD_ERROR_PARAM;
 swd_cmd_t **cmd, *tmpcmdq;
 cmd=calloc(count,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 int i;
 for (i=0;i<count;i++){
  cmd[i]->bits=1;
  cmd[i]->cmdtype=SWD_CMDTYPE_BITMISO;
  if (swd_cmd_queue_append(swdctx->cmdq, cmd[i])!=1){
   swd_cmd_queue_free_tail(cmd[0]); // if that fails, just clean the memory
   for (i=0;i<count;i++) free(cmd[i]);
   return SWD_ERROR_QUEUE;
  }
 }
 return count;
}

int swd_cmd_queue_append_mosi_nbit(swd_ctx_t *swdctx, int *data, int count){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
	if (data==NULL) return SWD_ERROR_NULLPOINTER;
	if (count<=0) return SWD_ERROR_PARAM;
	swd_cmd_t **cmd;
	cmd=calloc(count, sizeof(swd_cmt_t));
	if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
	int i;
	for (i=0;i<count;i++){
  cmd[i]->mosidata=data[i];
  cmd[i]->bits=1;
	 cmd[i]->cmdtype=SWD_CMDTYPE_BITMOSI;
  if (swd_cmd_queue_append(swdctx->cmdq, cmd[i])!=1){
   swd_cmd_queue_free_tail(cmd[0]); // if that fails, just clean the memory
   for (i=0;i<count;i++) free(cmd[i]);
   return SWD_ERROR_QUEUE;
  }
 }
 return count;
}

int swd_cmd_queue_append_mosi_parity(swd_ctx_t *swdctx, char *parity){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
	if (parity!=0 || parity!=1) return SWD_ERROR_PARAM;
 swd_cmd_t *cmd;
	cmd=calloc(1,sizeof(swd_cmd_t));
 if (swd==NULL) return SWD_ERROR_OUTOFMEM;
	cmd->parity=*parity;
	cmd->bits=1;
	cmd->cmdtype=SWD_CMDTYPE_MOSI_PARITY;
	return swd_cmd_queue_append(swdctx->cmdq, cmd);
}

int swd_cmd_queue_append_miso_parity(swd_ctx_t *swdctx, char *parity){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
	if (parity==NULL) return SWD_ERROR_NULLPOINTER;
 swd_cmd_t *cmd;
	cmd=calloc(1,sizeof(swd_cmd_t));
 if (swd==NULL) return SWD_ERROR_OUTOFMEM;
	parity=&cmd->parity;
	cmd->bits=1;
	cmd->cmdtype=SWD_CMDTYPE_MISO_PARITY;
	return swd_cmd_queue_append(swdctx->cmdq, cmd);
}




/* Append 32-bit element + 1-bit parity read operation into context cmd queue.*/
int swd_cmd_queue_append_miso_data(swd_ctx_t *swdctx, int *data){
 if (swdctx==NULL || data==NULL) return SWD_ERROR_NULLPOINTER;
 swd_cmd_t **cmd;
 int ret;
 cmd=calloc(2,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 //?cmd[0]->misodata=&data;
 cmd[0]->bits=32;
 cmd[0]->cmdtype=SWD_CMDTYPE_MISO;
 cmd[1]->parity;
 cmd[1]->bits=1;
 cmd[1]->cmdtype=SWD_CMDTYPE_PARMISO;
 ret=swd_cmd_queue_append(swdctx->cmdq, cmd[0]);
 if (ret<0) return ret;
 ret=swd_cmd_queue_append(swdctx->cmdq, cmd[1]);
 if (ret<0) return ret;
 return 2;
}

/* Append n-element read operation into context command queue. */
int swd_cmd_append_miso_ndata(swd_ctx_t *swdctx, int *data, int count){
 if (swdctx==NULL || data==NULL || count<=0) return SWD_ERROR_PARAM;
 int i;
 for (i=0;i<=count;i++) if (swd_cmd_append_read(swdctx, &data[i])<1) return SWD_ERROR_RESULT;
 return i;
}


int swd_cmd_append_mosi_data(swd_ctx_t *swdctx, int *data){
 if (swdctx==NULL || data==NULL) return SWD_ERROR_NULLPOINTER;
 swd_cmd_t **cmd;
 int ret;
 cmd=calloc(2,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 cmd[0]->mosidata=*data;
 cmd[0]->bits=32;
 cmd[0]->cmdtype=SWD_CMDTYPE_MOSI;
 cmd[1]->parity=swd_bin32_parity_even(*data);
 cmd[1]->bits=1;
 cmd[1]->cmdtype=SWD_CMDTYPE_MOSI;
 ret=swd_cmd_queue_append(swdctx->cmdq, cmd[0]);
 if (ret<0) return ret;
 ret=swd_cmd_queue_append(swdctx->cmdq, cmd[1]);
 if (ret<0) return ret;
 return 2;
}

int swd_cmd_append_mosi_ndata(swd_ctx_t *swdctx, int *data, int count){
 if (swdctx==NULL || data==NULL || count<1) return SWD_ERROR_PARAM;
 int i;
 for (i=0;i<count;i++) if (swd_cmd_append_write(swdctx, &data[i])<1) return SWD_ERROR_RESULT;
 return i;
}

/* Append Acnkowledge request from slave to the command queue. */
int swd_cmd_append_ack(swd_ctx_t *swdctx){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
 swd_cmd_t *cmd;
 cmd=calloc(1,sizeof(swd_cmd_t)); 
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;  
 cmd->bits=SWD_ACK_BITLEN;
 cmd->cmdtype=SWD_CMDTYPE_MISO_ACK;
 return swd_cmd_queue_append(swdctx->cmdq, cmd);
}

/* Append n-octet size control sequence to the context command queue. */
int swd_cmd_append_control(swd_ctx_t *swdctx, char *ctlmsg, int len){
 int elm, ret;
 swd_cmd_t **cmd;
 if (len<=0) return SWD_ERROR_PARAM;
 cmd=calloc(len,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 for (elm=0;elm<len;elm++){
  cmd[elm]->control=ctlmsg[elm];
  cmd[elm]->cmdtype=SWD_CMDTYPE_CONTROL;
  cmd[elm]->bits=sizeof(ctlmsg[elm]);
  ret=swd_cmd_queue_append(swdctx->cmdq, cmd[elm]); 
  if (ret<=1) return ret;
 }
 return elm;
}

int swd_cmd_append_swdpreset(swd_ctx_t *swdctx){
 return swd_cmd_append_control(swdctx, (char *)SWD_CMD_SWDPRESET, sizeof(SWD_CMD_SWDPRESET));
}

int swd_cmd_append_jtag2swd(swd_ctx_t *swdctx){
 return swd_cmd_append_control(swdctx, (char *)SWD_CMD_JTAG2SWD, sizeof(SWD_CMD_JTAG2SWD));
}

int swd_cmd_append_swd2jtag(swd_ctx_t *swdctx){
 return swd_cmd_append_control(swdctx, (char *)SWD_CMD_SWD2JTAG, sizeof(SWD_CMD_SWD2JTAG));
}

/* Check/set the bus direction to MOSI. */
int swd_bus_dir_mosi(swd_ctx_t *swdctx){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
 int res, cmdcnt=0;
 if ( swdctx->cmdq->prev==NULL || (swdctx->cmdq->cmdtype*SWD_CMDTYPE_MOSI<0) ) {
  res=swd_cmd_append_trnmosi(swdctx);
  if (res<0) return res;
  cmdcnt++;
 }
 return cmdnum;
}
 
/* Check/set the bus direction to MISO. */
int swd_bus_dir_miso(swd_ctx_t *swdctx){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
 int res, cmdcnt=0;
 if ( swdctx->cmdq->prev==NULL || (swdctx->cmdq->cmdtype*SWD_CMDTYPE_MISO<0) ) {
  res=swd_cmd_append_trnmiso(swdctx);
  if (res<0) return res;
  cmdcnt++;
 }
 return cmdnum;
}
 

int swd_send_request(swd_ctx_t *swdctx, char APnDP, char RnW, char addr){
 /* Verify function parameters.*/
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
 if (APnDP!=0 || APnDP!=1) return SWD_ERROR_APnDP;
 if (RnW!=0 || RnW!=1) return SWD_ERROR_RnW;
 if (addr<SWD_ADDR_MINVAL || addr>SWD_ADDR_MAXVAL) return SWD_ERROR_ADDR;

 /* Build request header content. */
 unsigned char reqhdr=0;
 int res, cmdcnt=0;
 reqhdr|=(((addr&(1<<2))?1:0)<<SWD_REQ_A2_BITNUM);
 reqhdr|=(((addr&(1<<3))?1:0)<<SWD_REQ_A3_BITNUM);
 reqhdr|=((APnDP?1:0)<<SWD_REQ_APnDP_BITNUM);
 reqhdr|=(((RnW?1:0)<<SWD_REQ_RnW_BITNUM));
 res=swd_bin8_parity_even(reqhdr);
 if (res<0 || res>1) return SWD_ERROR_PARITY;
 reqhdr|=(res<<SWD_REQ_PARITY_BITNUM);
 reqhdr|=(SWD_REQ_START_VAL<<SWD_REQ_START_BITNUM);
 reqhdr|=(SWD_REQ_STOP_VAL<<SWD_REQ_STOP_BITNUM);

 /* Bus direction must be MOSI. */
 res=swd_bus_dir_mosi(swdctx);
 if (res<0) return res;
 cmdcnt=+res;

 /* Append request command to the queue. */
 res=swd_cmd_append_request(swdctx, reqhdr);
 if (res<0) return res;
 cmdcnt+=res;

 return cmdnum; 
}


int swd_get_ack(swd_ctx_t *swdctx, char *ack){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
 if (ack==NULL) return SWD_ERROR_NULLPOINTER;
 int res;

 /* Bus direction must be MISO. */ 
 res=swd_bus_dir_miso(swdctx);
 if (res<0) return res;

 res=swd_cmd_append_ack(swdctx);
 if (res<0) return res;

 res=swd_transfer(swdctx);
 if (res<0) return res;

 /* If current command (ACK) was executed read the result. */
 if (cmdctx->cmdq->cmdtype==SWD_CMDTYPE_MISO_ACK && cmdctx->cmdq->done){
  ack=swdctx->cmdq.ack;
 } else return SWD_ERROR_ACK_READ;

	/* Return error if ACK response was unsuccessful. */
 switch(ack){
  case SWD_ACK_OK:
   return res;
		 break;
		case SWD_ACK_WAIT:
		 return SWD_ERROR_ACK_WAIT;
		 break;
		case SWD_ACK_FAULT:
		 return SWD_ERROR_ACK_FAULT;
		 break;	
		default:
		 return SWD_ERROR_ACK_UNKNOWN;	
 }

	/* We shouldn't get in here. */
 return SWD_ERROR_ACK;
}

int swd_send_data(swd_ctx_t *swdctx, int *data){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
	if (data==NULL) return SWD_ERROR_NULLPOINTER;
 int res, parity;

	res=swd_bus_mosi(swdctx);
	if (res<0) return ret;

	res=swd_cmd_append_mosi_data(swdctx, data);
 if (res<0) return ret;

 parity=swd_bit32_parity_even(*data);
	if (parity<0) return parity;
	res=swd_cmd_append_parity(swdctx, parity);
	if (res<0) return res;

	return 1;
}


int swd_read_idcode(swd_ctx_t *swdctx){
 
}

